from .utils import build_url


class UI:
    def __init__(self, server):
        self.server = server

    def url(self, endpoint, *args, **kwargs):
        return build_url(self.server, endpoint, *args, **kwargs)

    def dataset_url(self, id=None):
        url = self.url('datasets')
        if id:
            url += str(id)
        return url

    def annotate_url(self, id):
        return self.url('annotation', id)



import os
import subprocess
import time
from pathlib import Path

import requests

from .utils import browse
from .config import parse_config
from .sdk import SDK
from .domain.task import AnnotationTask

REMO_HOME = os.getenv('REMO_HOME', str(Path.home().joinpath('.remo')))
cfg_path = str(os.path.join(REMO_HOME, 'remo.json'))
config = parse_config(cfg_path)

logs = None
server = None

if config:

    def launch_server(open_browser=True):
        server_url = '{}:{}'.format(config.server, config.port)
        version_endpoint = '{}/version'.format(server_url)
        try:
            resp = requests.get(version_endpoint)
            print('Remo server is running:', resp.json())
            if open_browser:
                browse('{}:{}'.format(config.server, config.port))

            return True
        except:
            pass

        print('Launching Remo server ...')

        if not os.path.exists(config.conda_env):
            print("""
WARNING: Provided conda environment does not exist at {}

Failed to launch Remo server, please start it manually.

To fix config issue, please run init or update for Remo server (inside Remo conda env), like: 
python -m remo_app init
or
python -m remo_app update
""".format(config.conda_env))
            return False

        global logs
        global server
        logs = open(str(os.path.join(REMO_HOME, 'logs.txt')), 'w')

        path_args = ['python'] if os.name == 'nt' else ['bin', 'python']
        python_exe = os.path.join(config.conda_env, *path_args)

        server = subprocess.Popen(
            '{} -m remo_app'.format(python_exe),
            stdout=logs, stderr=logs,
            shell=True, universal_newlines=True,
        )

        time.sleep(5)
        retry = 5
        while retry > 0:
            print('Checking if server up ... ', retry)
            try:
                resp = requests.get(version_endpoint)
                print('Remo server is running:', resp.json())
                if open_browser:
                    browse(server_url)

                return True
            except:
                time.sleep(1)
            retry -= 1

        print('Failed to launch Remo server, please start it manually')
        return False


    launch_server(open_browser=False)

    sdk = SDK('{}:{}'.format(config.server, config.port), config.user_email, config.user_password)

    login = sdk.login
    datasets = sdk.datasets
    get_dataset = sdk.get_dataset
    search_images = sdk.search_images
    create_dataset = sdk.create_dataset


    def terminate_server():
        global logs
        global server
        print('Terminating Remo server')
        if server:
            server.terminate()
            logs.close()

else:
    print("""
    WARNING: Config not found in {}
    
    Please run init or update for Remo server (inside Remo conda env), like: 
    
    python -m remo_app init
    or
    python -m remo_app update
    
    """.format(cfg_path))



from .api import API
from .ui import UI
from .domain.interfaces import ISDK
from .domain.dataset import Dataset
from .domain.annotation_set import AnnotationSet
from .domain.task import AnnotationTask


class SDK(ISDK):
    def __init__(self, server, user_email=None, user_password=None):
        self.api = API(server)
        self.ui = UI(server)
        if user_email and user_password:
            self.api.login(user_email, user_password)

    # def search_images(self, search_terms_dictionary):
    #    result = self.api.search_images(search_terms_dictionary) 
    def search_images(self, cls, task):
        result = self.api.search_images(cls, task)

    def login(self, user_email, user_pwd):
        self.api.login(user_email, user_pwd)

    # ALR: do we need folder_id here?
    def create_dataset(self, name, local_files=[], paths_to_upload=[], urls=[], annotation_task=None,
                       folder_id=None, public=False) -> Dataset:

        # TODO: add documentation on annotation tasks and urls upload
        ''' Creates a dataset from an url or path

        Args:
            name: string, name of the Dataset
            files: list of paths of files
            urls: URL of images
            annotation_task: in case we are uploading annotations, specify the annotation task
            folder_id: 

        Returns: remo Dataset
        '''

        result = self.api.create_dataset(name, public)
        print(result)
        my_dataset = Dataset(self, **result)
        my_dataset.add_data(local_files, paths_to_upload, urls, annotation_task, folder_id)
        return my_dataset

    def datasets(self) -> []:
        """
        :return: list of datasets
        """
        resp = self.api.list_datasets()
        return [
            Dataset(self, id=dataset['id'], name=dataset['name'])
            for dataset in resp.get('results', [])
        ]

    def all_info_datasets(self, **kwargs) -> [Dataset]:
        result = self.api.all_info_datasets(**kwargs)
        return result

    def get_dataset(self, dataset_id) -> Dataset:
        '''
        Given a dataset id, returns the dataset
        '''
        result = self.api.get_dataset(dataset_id)
        return Dataset(self, **result)

    def add_data_to_dataset(self, dataset_id, local_files=[],
                            paths_to_upload=[], urls=[], annotation_task=None, folder_id=None):
        # JSONDecodeError: Expecting value: line 1 column 1 (char 0)
        '''
        Adds data to existing dataset
        
        Args:
            dataset_id: id of the desired dataset to extend (integer)
            
            local_files: list of files or directories. Function will scan for .png, .jpeg, .tiff and .jpg in the folders and sub-folders.
            
            paths_to_upload: list of files or directories. These files will be uploaded to the local disk.
                files supported: image files, annotation files and archive files.
                Annotation files: json, xml, csv. If annotation file is provided, you need to provide annotation task.
                Archive files: zip, tar, gzip. These files are unzipped, and then we scan for images, annotations and other archives. Support for nested archive files, image and annotation files in the same format supported elsewhere
            
            urls: list of urls pointing to downloadable target, which should be an archive file. The function will download the target of the URL - then we scan for archive files, unpack them and proceed as per Archive file section.
            
            annotation_task:
                object_detection = 'Object detection'. Supports Coco, Open Images, Pascal
                instance_segmentation = 'Instance segmentation'. Supports Coco
                image_classification = 'Image classification'. ImageNet
                
            folder_id: if there is a folder in the targer remo id, and you want to add images to a specific folder, you can specify it here.
            
        '''

        result = {}
        if len(local_files):
            if type(local_files) is not list:
                raise ValueError(
                    'Function parameter "paths_to_add" should be a list of file or directory paths, but instead is a ' + str(
                        type(local_files)))

            files_upload_result = self.api.upload_local_files(dataset_id, local_files, annotation_task, folder_id)
            result['files_link_result'] = files_upload_result

        if len(paths_to_upload):
            if type(paths_to_upload) is not list:
                raise ValueError(
                    'Function parameter "paths_to_upload" should be a list of file or directory paths, but instead is a ' + str(
                        type(paths_to_upload)))

            files_upload_result = self.api.bulk_upload_files(dataset_id=dataset_id,
                                                             files_to_upload=paths_to_upload,
                                                             annotation_task=annotation_task,
                                                             folder_id=folder_id)

            result['files_upload_result'] = files_upload_result

        if len(urls):
            if type(urls) is not list:
                raise ValueError(
                    'Function parameter "urls" should be a list of URLs, but instead is a ' + str(type(urls)))

            urls_upload_result = self.api.upload_urls(dataset_id=dataset_id,
                                                      urls=urls,
                                                      annotation_task=annotation_task,
                                                      folder_id=folder_id)

            print(urls_upload_result)
            result['urls_upload_result'] = urls_upload_result
        return result

    def annotation_sets(self, dataset_id):
        resp = self.api.list_annotation_sets(dataset_id)
        return [
            AnnotationSet(self,
                          id=annotation_set['id'],
                          name=annotation_set['name'],
                          task=AnnotationTask(annotation_set['task']['name']),
                          total_classes=annotation_set['statistics']['total_classes'])
            for annotation_set in resp.get('results', [])
        ]

    def ann_statistics(self, dataset_id):
        """        
        Args:
            dataset_id: int
        Returns: Lists annotation set statistics of the dataset
        """
        result = self.api.get_annotation_statistics(dataset_id)
        return result

    def list_dataset_images(self, dataset_id, folder_id=None, endpoint=None, **kwargs):
        if folder_id is not None:
            result = self.api.list_dataset_contents_by_folder(dataset_id, folder_id, **kwargs)
        else:
            result = self.api.list_dataset_contents(dataset_id, **kwargs)

        # print('Next:', result.get('next'))
        images = []
        for entry in result.get('results', []):
            name = entry.get('name')
            images.append(name)

        return images

    def export_annotations(self, annotation_set_id, annotation_format=None):
        """
        Exports annotation set 
        Args:
            annotation_set_id: int
            annotation_format: choose format from this list ['json', 'coco'], default = 'json'
        Returns: annotations
        """
        args = [annotation_set_id]
        if annotation_format:
            args.append(annotation_format)
        return self.api.export_annotations(*args)

    def show_images(self, dataset_id, image_id):
        return self.api.show_images(dataset_id, image_id)



from abc import ABCMeta, abstractmethod


class ISDK(metaclass=ABCMeta):

    @abstractmethod
    def datasets(self) -> []:
        """
        :return: list of datasets
        """
        raise NotImplementedError

    @abstractmethod
    def annotation_sets(self, dataset_id):
        """
        :return: list of annotation sets for giving dataset
        """
        raise NotImplementedError



from .interfaces import ISDK
from .. import utils
from io import BytesIO
import requests


class Image:
    id = None
    dataset = None
    path_to_image = None
    annotation_sets = []

    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.dataset = kwargs.get('dataset')
        self.path_to_image = kwargs.get('path')
        # print(id, dataset,path_to_image, '\n')


class Dataset:
    def __init__(self, sdk: ISDK, **kwargs):
        self.sdk = sdk
        self.id = kwargs.get('id')
        self.name = kwargs.get('name')
        self.images = []
        self._annotation_sets = []
        self.default_annotation_set = None

    def __str__(self):
        return "Dataset {id} - '{name}'".format(id=self.id, name=self.name)

    def __repr__(self):
        return self.__str__()

    def annotation_sets(self):
        return self.sdk.annotation_sets(self.id)

    def initialise_images(self):
        list_of_images = self.list_images()
        for i_image in list_of_images:
            my_image = Image(id=None, path=i_image, dataset=self.name)
            print(i_image, self.name)
            self.images.append(my_image)

    def list_images(self, folder_id=None, **kwargs):
        return self.sdk.list_dataset_images(self.id, folder_id=None, **kwargs)

    def add_data(self, local_files=[], paths_to_upload=[], urls=[], annotation_task=None, folder_id=None):
        """
        Adds images and optionally annotations to a Dataset

        Longer Description

        Args:
            files: 
            urls: 
            annotation_task: 
            folder_id:

        Returns:


        Raises:

        """

        return self.sdk.add_data_to_dataset(dataset_id=self.id,
                                            local_files=local_files,
                                            paths_to_upload=paths_to_upload,
                                            urls=urls,
                                            annotation_task=annotation_task,
                                            folder_id=folder_id)

    def fetch(self):
        dataset = self.sdk.get_dataset(self.id)
        self.__dict__.update(dataset.__dict__)

    def browse(self):
        utils.browse(self.sdk.ui.dataset_url(self.id))

    def annotate(self, ann_id):
        # TODO: select by annotation task
        utils.browse(self.sdk.ui.annotate_url(ann_id))
        # if len(self.annotation_sets) > 0:

        #    utils.browse(self.sdk.ui.annotate_url(self.annotation_sets[0]))
        # else:
        #    print("No annotation sets in dataset " + self.name)

    def search(self, **kwargs):
        pass

    def get_images(self, cls=None, tag=None):
        # TODO: add class and tags
        # dataset_details = self.sdk.all_info_datasets()
        # dataset_info = None
        # for res in dataset_details['results']:
        #    if res['id'] == self.id:
        #        dataset_info = res
        # url_list = []
        # image_thumbnails = dataset_info.get('image_thumbnails')
        # for i in range(len(image_thumbnails)):
        #    url_ = image_thumbnails[i]['image']
        #    url_list.append(url_)
        # return url_list
        return 0

    def show_images(self, image_id=17, cls=None, tag=None):
        return self.sdk.show_images(self.id, image_id)
        # TODO: redirect to ui with endpoints
        # image_urls = self.get_images(cls=None, tag=None)
        # imgs = []
        # for url in image_urls:
        #    bytes_ = requests.get(url).content
        #    rawIO = BytesIO(bytes_)
        #    imgs.append(rawIO)
        # return imgs

    def show_objects(self, cls, tag):
        pass



import http
import os
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timedelta
import filetype
import requests
from .domain.task import AnnotationTask
from .utils import FileResolver, build_url, browse


class UploadStatus:

    def __init__(self, total_count: int):
        self.total_count = total_count
        self.current_count = 0
        self.start = datetime.now()
        self.reported_progress = 0

    def update(self, count: int):
        self.current_count += count

    def progress(self):
        percentage = int(self.current_count / self.total_count * 100)
        if percentage > self.reported_progress:
            elapsed = (datetime.now() - self.start).seconds + 1e-3
            avg_speed = self.current_count / elapsed
            eta = timedelta(seconds=(self.total_count - self.current_count) / avg_speed)
            print(
                'Progress {}% - {}/{} - elapsed {} - speed: {} img / s, ETA: {}'.format(percentage, self.current_count,
                                                                                        self.total_count,
                                                                                        timedelta(seconds=elapsed),
                                                                                        "%.2f" % avg_speed, eta))
        self.reported_progress = percentage


class NotAuthenticatedError(Exception):
    """ Not Authenticated run-time error. """

    def __init__(self, *args, **kwargs):
        pass


class API:
    def __init__(self, server):
        self.server = server.rstrip('/')
        self.token = None

    def url(self, endpoint, **kwargs):
        return build_url(self.server, endpoint, **kwargs)

    def is_authenticated(self):
        return self.token is not None

    def auth_header(self):
        if not self.is_authenticated():
            raise NotAuthenticatedError

        return {'Authorization': 'Token {}'.format(self.token)}

    def post(self, *args, **kwargs):
        return requests.post(*args, headers=self.auth_header(), **kwargs)

    def get(self, *args, **kwargs):

        return requests.get(*args, headers=self.auth_header(), **kwargs)

    def login(self, user_email, user_pwd):
        try:
            resp = requests.post(self.url('/api/rest-auth/login'),
                                 data={"password": user_pwd, "email": user_email})
        except requests.exceptions.ConnectionError:
            print('ERROR: Failed connect to server')
            return

        if resp.status_code != http.HTTPStatus.OK:
            raise Exception(resp.json())

        self.token = resp.json().get('key')

    def create_dataset(self, name, public=False):
        return self.post(self.url('/api/dataset'),
                         json={"name": name, "is_public": public}).json()

    def upload_file(self, dataset_id, path, annotation_task=None, folder_id=None):
        name = os.path.basename(path)
        files = {'files': (name, open(path, 'rb'), filetype.guess_mime(path))}
        data = {}
        if annotation_task:
            data['annotation_task'] = annotation_task.value

        url = self.url('/api/dataset/{}/upload'.format(dataset_id), folder_id=folder_id)
        return self.post(url, files=files, data=data).json()

    # TODO: fix progress to include both local files and uploads
    def upload_files(self, dataset_id: int,
                     files_to_upload: list = [],
                     annotation_task: AnnotationTask = None,
                     folder_id: int = None,
                     status: UploadStatus = None):

        files = [('files', (os.path.basename(path), open(path, 'rb'), filetype.guess_mime(path))) for path in
                 files_to_upload]
        data = {}
        if annotation_task:
            data['annotation_task'] = annotation_task.value

        url = self.url('/api/dataset/{}/upload'.format(dataset_id), folder_id=folder_id)

        r = self.post(url, files=files, data=data)

        if r.status_code != http.HTTPStatus.OK:
            print('Possible Error - Response:', r.text, 'files:', files_to_upload)

        status.update(len(files))
        status.progress()
        return r.json()

    # TODO: fix progress to include both local files and uploads
    def bulk_upload_files(self, dataset_id: int, files_to_upload: list,
                          annotation_task: AnnotationTask = None,
                          folder_id: int = None):

        # files to upload
        files = FileResolver(files_to_upload, annotation_task is not None).resolve()
        groups = self.split_files_by_size(files)
        status = UploadStatus(len(files))
        with ThreadPoolExecutor(1) as ex:
            res = ex.map(lambda bulk: self.upload_files(dataset_id, bulk, annotation_task, folder_id, status),
                         groups)

        results = res
        return results

    def chunks(self, my_list, chunk_size=2000):
        groups = []
        """Yield successive n-sized chunks from l."""
        for i in range(0, len(my_list), chunk_size):
            groups.append(my_list[i:i + chunk_size])

        return groups

    def split_files_by_size(self, files: list) -> list:
        groups = []
        bulk = []
        total_size = 0
        bulk_size = 8 * 1024 * 1024
        for path in files:
            total_size += os.path.getsize(path)
            bulk.append(path)
            if total_size >= bulk_size:
                groups.append(bulk)
                bulk = []
                total_size = 0

        if len(bulk):
            groups.append(bulk)
        return groups

    def upload_local_files(self, dataset_id, local_files, annotation_task=None, folder_id=None):
        payload = {"local_files": local_files}
        if annotation_task:
            payload['annotation_task'] = annotation_task.value
        if folder_id:
            payload['folder_id'] = folder_id

        url = self.url('/api/dataset/{}/upload'.format(dataset_id))
        return self.post(url, json=payload).json()

    def upload_urls(self, dataset_id, urls, annotation_task=None, folder_id=None):
        payload = {"urls": urls}
        if annotation_task:
            payload['annotation_task'] = annotation_task.value
        if folder_id:
            payload['folder_id'] = folder_id

        url = self.url('/api/dataset/{}/upload'.format(dataset_id))
        return self.post(url, json=payload).json()

    def list_annotation_sets(self, dataset_id: int):
        url = self.url('/api/v1/ui/datasets/{}/annotation-sets/'.format(dataset_id))
        return self.get(url).json()

    def list_datasets(self):
        """
        :return: list of datasets
        """

        url = self.url('/api/v1/ui/datasets')
        return self.get(url).json()

    # MC: it's export_annotations()
    # def get_annotation_set(self, ann_set_id, endpoint=None):
    #    url = None
    #    if not endpoint:
    #        return get_ann_set(ann_set_id)
    #    else:
    #        return self.get(url).json()

    def get_annotation_statistics(self, dataset_id):
        """
        Args:
            dataset_id: int
        Returns: annotation statistics
        """

        url = self.url('/api/v1/ui/datasets/{}/annotation-sets'.format(dataset_id))
        return self.get(url).json()

    def get_dataset(self, id):
        url = self.url('/api/dataset/{}'.format(id))
        return self.get(url).json()

    def all_info_datasets(self, endpoint=None, **kwargs):
        if endpoint:
            url = self.url('/api/dataset', **kwargs)
            return self.get(url).json()
        else:
            return dset_info()

    def show_images(self, dataset_id, image_id):
        url = self.url('/image/{}?datasetId={}').format(image_id, dataset_id)
        return browse(url)

    def search_images(self, cls=None, task=None):
        if cls:
            url = self.url('/api/v1/ui/search/?classes={}&limit=5').format(cls)
        else:

            url = self.url('/api/v1/ui/search/?tasks=Instance%20segmentation&limit=15')
            # url = self.url('/api/v1/ui/search/?tasks=Image%20classification&limit=15').format(task)
        return browse(url)

    def get_images(self, dataset_id, image_id):
        # MC: dowdloads the images
        # TODO: get using dataset_id
        url = self.url('/media/dataset_images/')
        return self.get(url).json()

    def list_dataset_contents(self, dataset_id, **kwargs):
        url = self.url('/api/v1/ui/datasets/{}/images'.format(dataset_id), **kwargs)
        return self.get(url).json()

    def list_dataset_contents_by_folder(self, dataset_id, folder_id, **kwargs):
        url = self.url('/api/user-dataset/{}/contents/{}'.format(dataset_id, folder_id), **kwargs)
        return self.get(url).json()

    def export_annotations(self, annotation_set_id, annotation_format='json'):
        """
        Args:
            annotation_set_id: int
            annotation_format: can be one of ['json', 'coco'], default='json'
        Returns: annotations
        """
        url = self.url(
            'api/v1/ui/annotation-sets/{}/export/?annotation-format={}'.format(annotation_set_id, annotation_format))
        return self.get(url).json()
